# A tagger that associates words with their most likely tags in a
# training corpus, using only leading or trailing substrings of each
# word.

### CLASSDEF

setClass("AffixTagger",
    representation(
      affix.length = "integer", # subtr size, neg=suffix
      min.stem.length = "integer", # use backoff for stems short than this
      cutoff  = "integer", # exclude most likely tag for a context if
                           # it occurs less than cutoff times
      "WrappedSequentialBackoffTagger" # extends the sbt virtual class
    )
)

# TODO: validity

### GENERICS
setGeneric("AffixTagger",
  function (trainset, affix.length, min.stem.length, cutoff, backoff) 
    standardGeneric("AffixTagger"))
setGeneric("affix.length",
  function (object) standardGeneric("affix.length"))
setGeneric("min.stem.length",
  function (object) standardGeneric("min.stem.length"))

### CONSTRUCTION

# The AT constructors are methods because we can construct ATs from
# a variety of training set objects.

# Constructor for "list" trainset.  The expected format is a three
# level nested list of the form (sentence_1, sentence_2, ... sentence_N)
# where sentence_n = ((tag_1, token_1) ... (tag_N, token_n)).

# Constructor for nested-list training data
setMethod("AffixTagger", signature(trainset = "list"),
  function(trainset, affix.length, min.stem.length, cutoff, backoff) {
    if (! is(backoff, "SequentialBackoffTagger"))
      if (is.na(backoff))
        backoff <- NULL

    if (! is.numeric(affix.length))
      stop("affix.length must be numeric")
    if (! is.numeric(min.stem.length) | min.stem.length < 0)
      stop("min.stem.length must be a positive number")
    if (! is.numeric(cutoff) | cutoff < 1)
      stop("cutoff must be a positive numeric")
    if (! (is.null(backoff)  | 
           is(backoff, "SequentialBackoffTagger")))
      stop("backoff must be SequentialBackoffTagger or NULL")
    if (! goodTaggedSentenceList(trainset))
      stop("Training list format is incorrect")

    cutoff <- as.integer(cutoff)
    affix.length <- as.integer(affix.length)
    min.stem.length <- as.integer(min.stem.length)

    ptr <- .Call("construct_affix_tagger", trainset, affix.length,
                 min.stem.length, cutoff)
    self <- new("AffixTagger", pointer = ptr, affix.length = affix.length, 
                min.stem.length = min.stem.length, cutoff = cutoff)
    self <- initSequentialBackoffTagger(self, backoff)
    return(self)
  }
)

# TODO: Constructors for structured (sentence-chunked) docs

### ACCESSORS
setMethod("cutoff", "AffixTagger", function (object) object@cutoff)
setMethod("affix.length", "AffixTagger",
  function (object) object@affix.length)
setMethod("min.stem.length", "AffixTagger", 
  function (object) object@min.stem.length)

### SHOW

setMethod("show", "AffixTagger",
  function (object)
    cat(paste("An affix tagger with affix length ", affix.length(object),
              ", minimum stem length ", min.stem.length(object),
              ", and cutoff ", cutoff(object), ".\n", sep="")))
