# A tagger that associates words with their most likely tags in a
# training corpus, using information from the n words preceeding the
# given word.

### CLASSREF

setClass("NgramTagger",
    representation(
      n = "integer", # the n-back to include in the context
      cutoff  = "integer", # exclude most likely tag for a context if
                           # it occurs less than cutoff times
      "WrappedSequentialBackoffTagger" # extends the sbt virtual class
    ),
)

# TODO: validity

### GENERICS

setGeneric("NgramTagger", 
  function (n, trainset, cutoff, backoff) standardGeneric("NgramTagger"))
setGeneric("n", function (object) standardGeneric("n"))
setGeneric("cutoff", function (object) standardGeneric("cutoff"))

### CONSTRUCTION

# The NGT constructors are methods because we can construct NGTs from
# a variety of training set objects.

# Constructor for "list" trainset.  The expected format is a three
# level nested list of the form (sentence_1, sentence_2, ... sentence_N)
# where sentence_n = ((token_1, tag_1) ... (token_N, tag_n)).
setMethod("NgramTagger", signature(n = "numeric", trainset = "list"),
  function(n, trainset, cutoff, backoff) {
    if (! is(backoff, "SequentialBackoffTagger"))
      if (is.na(backoff))
        backoff <- NULL

    if (n < 1)
      stop("n must be at least 1")
    if (! is.numeric(cutoff) | cutoff < 1)
      stop("cutoff must be a positive numeric")
    if (! (is.null(backoff)  | 
           is(backoff, "SequentialBackoffTagger")))
      stop("backoff must be SequentialBackoffTagger or NULL")
    if (! goodTaggedSentenceList(trainset))
      stop("Training list format is incorrect")

    n <- as.integer(n)
    cutoff <- as.integer(cutoff)

    ptr <- .Call("construct_ngram_tagger", n, trainset, cutoff)
    self <- new("NgramTagger", pointer = ptr, n = n, cutoff = cutoff)
    self <- initSequentialBackoffTagger(self, backoff)
    return(self)
  }
)

# TODO: Constructors for structured (sentence-chunked) docs

### ACCESSORS

# Get the number of tokens included in the tagger's context
setMethod("n", "NgramTagger", function (object) object@n)

# Get the tagger's cutoff (the number of times a  tag has to occur in the
# training corpus to be included)
setMethod("cutoff", "NgramTagger", function (object) object@cutoff)

### SPECIAL CASES

# Abbreviations for common ngram taggers
setClass("UnigramTagger", representation("NgramTagger"))

UnigramTagger <- function (trainset, cutoff = 1, backoff = NA)
{
  return (NgramTagger(1, trainset, cutoff, backoff))
}

setClass("BigramTagger", representation("NgramTagger"))

BigramTagger <- function (trainset, cutoff = 1, backoff = NA)
{
  return (NgramTagger(2, trainset, cutoff, backoff))
}

setClass("TrigramTagger", representation("NgramTagger"))

TrigramTagger <- function (trainset, cutoff = 1, backoff = NA)
{
  return (NgramTagger(3, trainset, cutoff, backoff))
}

### SHOW

setMethod("show", "NgramTagger",
  function (object)
    cat("An Ngram Tagger of order ", n(object),
        " with a minimum frequency cutoff of ",
        cutoff(object), ".\n"))
