# RegexObjects are simple wrappers around the python equivalent
setClass("RegexObject", representation(pointer="externalptr"))

### GENERICS

setGeneric("re.match",
  function (pattern, string, pos=NA, endpos=NA) standardGeneric("re.match")) 
setGeneric("re.search",
  function (pattern, string, pos=NA, endpos=NA) standardGeneric("re.search")) 
setGeneric("re.split",
  function (pattern, string, maxsplit=0) standardGeneric("re.split"))

setGeneric("re.findall",
  function (pattern, string, pos=NA, endpos=NA)
    standardGeneric("re.findall"))

setGeneric("re.sub",
  function (pattern, repl, string, count=0) standardGeneric("re.sub"))

setGeneric("re.subn",
  function (pattern, repl, string, count = 0) standardGeneric("re.subn"))

setGeneric("re.flags", function (pattern) standardGeneric("re.flags"))

setGeneric("re.groupindex",
  function (pattern) standardGeneric("re.groupindex"))

setGeneric("re.pattern",
  function (pattern) standardGeneric("re.pattern"))

# A helper
fixpos <- function(pos, length) {
  if (is.na(pos) || (is.numeric(pos) && all(pos >= 0))) {
    pos <- rep(pos, length.out=length)
    pos <- as.integer(pos)
    return(pos)
  }

  error("Position must be 0 or greater")
}

# The following methods all behave like the python methods that they
# wrap.  See http://docs.python.org/lib/re-objects.html and
# http://docs.python.org/lib/node46.html for details.  The only major
# difference is that these methods operate on vectors, where
# appropriate, applying the basic python operation to each iterm in
# the vector, and recycle accompanying arguments as necessary to
# obtain this result.  For example, match.re applies the python match
# routine to every character string in the parameter "string", even if
# it is vector valued and recyles pos and endpos as necessary to do
# so.  Other differences are noted in comments preceeding the
# methods, when they exist.

# Returns NA if no match
setMethod("re.match", signature(pattern="RegexObject", string="character"),
  function(pattern, string, pos=NA, endpos=NA) {
    if (is.na(pos) && ! is.na(endpos))
      error("Cannot specify endpos without pos")

    pos <- fixpos(pos, length(string))
    endpos <- fixpos(endpos, length(string))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      ptr <- .Call("match_RegexObject", pattern@pointer, string[i],
                   pos[i], endpos[i])
      if (is.null(ptr))
        result[[i]] <- NA
      else
        result[[i]] <- new("MatchObject", pointer = ptr)
    }

    if (length(result) == 1)
      return(result[[1]])

    return(result)
  }
)

# Return NA if no match
setMethod("re.search", signature(pattern="RegexObject",string="character"),
  function (pattern, string, pos=NA, endpos=NA) {
    if (is.na(pos) && ! is.na(endpos))
      error("Cannot specify endpos without pos")

    pos <- fixpos(pos, length(string))
    endpos <- fixpos(endpos, length(string))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      ptr <- .Call("search_RegexObject", pattern@pointer, string[i],
                   pos[i], endpos[i])
      if (is.null(ptr))
        result[[i]] <- NA
      else
        result[[i]] <- new("MatchObject", pointer = ptr)
    }

    if (length(result) == 1)
      return(result[[1]])

    return(result)
  }
)

setMethod("re.split", signature(pattern="RegexObject", string="character"),
  function (pattern, string, maxsplit=0) {
    if (! is.numeric(maxsplit) || (! all(maxsplit >= 0)))
      error("maxsplit elements must be integers >= 0")

    maxsplit <- as.integer(rep(maxsplit, length.out=length(string)))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      result[[i]] <- .Call("split_RegexObject", pattern@pointer, string[i],
                           maxsplit[i])
    }

    if (length(result) == 1)
      return (result[[1]])

    return(result)
  }
)

setMethod("re.findall", signature(pattern="RegexObject",string="character"),
  function (pattern, string, pos=NA, endpos=NA) {
    if (is.na(pos) && ! is.na(endpos))
      error("Cannot specify endpos without pos")

    pos <- fixpos(pos, length(string))
    endpos <- fixpos(endpos, length(string))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      result[[i]] <- .Call("findall_RegexObject", pattern@pointer,
                           string[i], pos[i], endpos[i])
    }

    if (length(result) == 1)
      return(result[[1]])

    return(result)
  }
)

# NOTE: Vector length is set by string here, repl is recycled as need.
# TODO(tricky): Allow function arguments to repl
setMethod("re.sub", signature(pattern="RegexObject", repl="character",
                              string="character"),
  function (pattern, repl, string, count=0) {
    if (! is.numeric(count) || (! all(count >= 0)))
      error("count elements must be integers >= 0")

    count <- as.integer(rep(count, length.out=length(string)))
    repl <- rep(repl, length.out=length(string))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      result[[i]] <- .Call("sub_RegexObject", pattern@pointer,
                           repl[i], string[i], count[i])
    }

    if (length(result) == 1)
      return (result[[1]])

    return(result)
  }
)

# NOTE: Vector length is set by string here, repl is recycled as need.
# TODO(tricky): Allow function arguments to repl
setMethod("re.subn", signature(pattern="RegexObject", repl="character",
                              string="character"),
  function (pattern, repl, string, count=0) {
    if (! is.numeric(count) || (! all(count >= 0)))
      error("count elements must be integers >= 0")

    count <- as.integer(rep(count, length.out=length(string)))
    repl <- rep(repl, length.out=length(string))
    result <- vector("list", length(string))

    for (i in 1:length(result)) {
      result[[i]] <- .Call("subn_RegexObject", pattern@pointer,
                           repl[i], string[i], count[i])
    }

    if (length(result) == 1)
      return (result[[1]])

    return(result)
  }
)
# Rather than simply return an integer like the python equivalent,
# this returns a character vector of flag names.
setMethod("re.flags", signature(pattern="RegexObject"),
  function (pattern) {
    flagstr <- c("IGNORECASE", "LOCALE", "MULTILINE", "DOTALL",
                 "UNICODE", "VERBOSE")
    flags <- character(0)

    flagint <- .Call("flags_RegexObject", pattern@pointer)

    for (i in 6:1) {
      if (flagint >= 2^i) {
        flagint <- flagint - 2^i
        flags <- c(flagstr[i], flags)
      }
    }

    return (flags)
  }
)

setMethod("re.pattern", signature(pattern="RegexObject"),
  function (pattern) return(.Call("pattern_RegexObject", pattern@pointer)))

# Construct a RegexObject
re.compile <- function(pattern, ignorecase=FALSE, locale=FALSE, 
                       multiline=FALSE, dotall=FALSE, unicode=FALSE,
                       verbose=FALSE)
{
  flagint <- 0
  if (ignorecase) flagint <- flagint + 2
  if (locale) flagint <- flagint + 4
  if (multiline) flagint <- flagint + 8
  if (dotall) flagint <- flagint + 16
  if (unicode) flagint <- flagint + 32
  if (verbose) flagint <- flagint + 64

  ptr <- .Call("construct_regex_object", pattern, as.integer(flagint))
  return(new("RegexObject", pointer = ptr))
}

setMethod("show", "RegexObject",
  function(object) {
    cat("RegexObject ", re.pattern(object), "\n")
  }
)
