# A tagger that attempts to associate tokens with tags given a set of
# regex-tag pairs, given as a nested list, where each sublist is a
# regex/tag pair.  Also takes an optional backoff tagger, which is
# used to tag tokens for which the current tagger always agrees with
# the backoff, or for which it provides no tag.  This class wraps an
# underlying nltk regexptagger.

### CLASSDEF

setClass("RegexTagger",
    representation(
      pattern = "list",     # a list or regex-tag pairs
      "WrappedSequentialBackoffTagger" # extends the sbt virtual class
    )
)

### VALIDITY

"isValidRegexTagger" <-
# A valid RegexTagger should have a pattern vector which is a
# character vector of even length
function(object) {
  pattern <- pattern(object)
  for (i in 1:length(pattern)) {
    if (length(pattern[[i]]) != 2)
      return("patterns must be pairs")
  }

  return(TRUE)
}
setValidity("RegexTagger", isValidRegexTagger)

### CONSTRUCTION

# A constructor for RegexTaggers
RegexTagger <- function(pattern = NULL, backoff = NULL) 
{
  if (is.null(pattern)) {
    pattern <- list(
      list('^-?[0-9]+(.[0-9]+)?$', 'CD'),   # cardinal numbers
      list('^(The|the|A|a|An|an)$', 'AT'),  # articles
      list('.*able$', 'JJ'),                # adjectives
      list('.*ness$', 'NN'),                # nouns formed from adjectives
      list('.*ly$', 'RB'),                  # adverbs
      list('.*ing$', 'VBG'),                # gerunds
      list('.*ed$', 'VBD'),                 # past tense verbs
      list('.*es$', 'VBZ'),                 # 3rd singular present verbs
      list('.*ould$', 'MD'),                # modals
      list('.*\'s', 'NN$'),                 # possessive nouns
      list('.*s$', 'NNS')                  # plural nouns
    )
    # Will never get to a backoff, so feel free to overwrite
    backoff <- DefaultTagger("NN");
  }

  ptr <- .Call("construct_regexp_tagger",pattern)
  self <- new("RegexTagger", pattern = pattern, pointer = ptr)
  self <- initSequentialBackoffTagger(self, backoff)

  return(self)
}
### ACCESSORS

# Access the tag of the given RegexTagger
setMethod("pattern", "RegexTagger", 
  function(object) {
    return(object@pattern)
  }
)


### SHOW

setMethod("show", "RegexTagger",
  function (object)
    cat("A regular expression tagger using the pattern:\n",
        format(pattern(object)), "\n"))
