# A virtual class for taggers that tag tokens sequentially, left to
# right with optional backoff tagger for tokens they can't tag.
# Extends the base tagger class.  Contains an optional (XXX reference
# to) a backoff tagger to apply, if it can't find a tag for a given
# token.

### CLASSDEF

# NOTE: We make backoff of type ANY for convenience.  Would use
# SequentialBackoffTagger but it is nice to leave it NULL when not
# used.  Extenders should always use initSequentialBackoffTagger to
# set the backoff which takes care of type checking.
setClass("SequentialBackoffTagger", 
  representation(backoff="ANY", "Tagger", "VIRTUAL"))

### GENERICS
setGeneric("choose.tag",
  function (tokens, index, history, tagger)
    standardGeneric("choose.tag"))
setGeneric("backoff", function (object) standardGeneric("backoff"))
setGeneric("tag.one",
  function (tokens, index, history, tagger)
    standardGeneric("tag.one"))

### UTILITIES FOR EXTENDING CLASSES

# Extending classes must implement the method:
#
# setMethod("choose.tag", signature(token="character",
# index="integer", history="character", tagger="MySBT") ...
#
# where token is the token to tag, i is the token's position in the
# token stream, and history is the set of tags applied to the tokens
# before this one.  This method should return NULL if no tag is
# available.  Do NOT consult the backoff tagger if no tag is found.
setMethod("choose.tag", signature(tokens="character", index="integer",
                                  history="character",
                                  tagger="SequentialBackoffTagger"),
  function (tokens, index, history, tagger) {
    stop("choose.tag must be implemented by derived class")
  }
)

# Take care of the stuff that needs done when creating any sbt.
# Assumes self's backoff slot is empty
initSequentialBackoffTagger <- function (self, backoff=NULL)
{
  if (! is(self, "SequentialBackoffTagger"))
    stop("This function only initializes SequentialBackoffTagger objects")
  if (! is.null(backoff)) {
    if (! is(backoff, "SequentialBackoffTagger"))
      stop("Backoff tagger must be of type SequentialBackoffTagger")
    self@backoff <- backoff
  }

  return (self)
}
  
### ACCESSORS

# Access the backoff tagger of the given SequentialBackoffTagger
setMethod("backoff", "SequentialBackoffTagger",
  function(object) {
    return(object@backoff)
  }
)

### INTERFACE UTILITY METHODS

# Tag a single token, using a backoff tagger if necessary.
setMethod("tag.one", signature(tokens="character", index="integer",
                               history="character",
                               tagger="SequentialBackoffTagger"),
  function (tokens, index, history, tagger) {
    tag <- NULL
    while (! is.null(tagger)) {
      tag <- choose.tag(tokens, index, history, tagger)
      if (! is.null(tag))
        break;
      tagger <- backoff(tagger)
    }

    if (is.null(tag))
      tag <- ""

    return(tag)
  }
)

# Tag a sequence of tokens.
setMethod("tag", signature(object="character", 
                           tagger="SequentialBackoffTagger",
                           tokenizer="missing",
                           index="missing"),
  function (object, tagger, tokenizer, index) {
    if (is(object, "TextDocument"))
      object <- tokenize(object)
    tags <- character()
    lapply(1:length(object), function (i) {
      tags <<- c(tags, tag.one(object, i, tags, tagger))
    })
    return(tags)
  }
)
