
### CLASSDEF

setClass("TaggedCorpusView",
  representation("TaggedView", "TokenizedCorpusView",
                 uniqueTags = reftype("character"),
                 uniqueTokenTags = reftype("TaggedToken"),
                 tagFreqTable = reftype("table"),
                 tokenTagFreqTable = reftype("table"),
                 documentTagMatrix = reftype("TermDocMatrix"),
                 documentTokenTagMatrix = reftype("TermDocMatrix")))

### GENERICS

setGeneric("documentTaggedTokenMatrix",
  function (object, weight) standardGeneric("documentTaggedTokenMatrix"))
setGeneric("documentTagMatrix",
  function (object, weight) standardGeneric("documentTagMatrix"))
### CONSTRUCTION

TaggedCorpusView <- function (view, tagger, transforms=list(),
                              documentFilters=list()) {
  # Tokenized bits
  documentViews <- new("NULLlist")
  tokenFreqTable <- new("NULLtable")
  uniqueTokens <- new("NULLcharacter")
  documentTokenMatrix <- new("NULLTermDocMatrix")

  # Tagged bits
  uniqueTags <- new("NULLcharacter")
  uniqueTokenTags <- new("NULLTaggedToken")
  tagFreqTable <- new("NULLtable")
  tokenTagFreqTable <- new("NULLtable")
  documentTagMatrix <- new("NULLTermDocMatrix")
  documentTokenTagMatrix <- new("NULLTermDocMatrix")

  # XXX for now, don't bother with this because it seems like a
  # sub-optimal approach in most cases, unlike the dv-level situation.
  # If this is a simple tagging we can save some computation
  #if (length(transforms) == 0 && length(documentFilters) == 0 &&
  #    .keepComputed) {
  #  uniqueTokens <- ref(view, "uniqueTokens")
  #  tokenFreqTable <- ref(view, "tokenFreqTable")
  #  documentTokenMatrix <- ref(view, "documentTokenMatrix")
  #}

  # Update transforms and doc-filters
  transforms = c(tokenTransforms(view), transforms)
  documentFilters = c(documentFilters(view), documentFilters)

  res <- new("TaggedCorpusView",
             corpus=corpus(view),
             keepComputed=.keepComputed,
             tokenizer=tokenizer(view),
             transforms=transforms,
             documentFilters=documentFilters,
             documentViews=documentViews,
             keepDocumentViews=.keepDocumentViews,
             tokenFreqTable=tokenFreqTable,
             uniqueTokens=uniqueTokens,
             documentTokenMatrix=documentTokenMatrix,
             tagger=tagger,
             uniqueTags=uniqueTags,
             uniqueTokenTags=uniqueTokenTags,
             tagFreqTable=tagFreqTable,
             tokenTagFreqTable=tokenTagFreqTable,
             documentTagMatrix=documentTagMatrix,
             documentTokenTagMatrix=documentTokenTagMatrix)

  registerView(corpus(view), res)
  return(res)
}

### CALLBACK methods

setMethod("clearComputed", "TaggedCorpusView",
  function (object) {
    ref(object, "tokenFreqTable") <- new("NULLtable")
    ref(object, "uniqueTokens") <- new("NULLcharacter")
    ref(object, "documentTokenMatrix") <- new("NULLTermDocMatrix")

    ref(object, "uniqueTags") <- new("NULLcharacter")
    ref(object, "uniqueTokenTags") <- new("NULLTaggedToken")
    ref(object, "tagFreqTable") <- new("NULLtable")
    ref(object, "tokenTagFreqTable") <- new("NULLtable")
    ref(object, "documentTagMatrix") <- new("NULLTermDocMatrix")
    ref(object, "documentTokenTagMatrix") <- new("NULLTermDocMatrix")
  })

# inherits updateView

### ACCESSORS

setMethod("documentViews", "TaggedCorpusView",
  function (object) {
    if (is(ref(object, "documentViews"), "NULLlist")) {
      proto <- TaggedDocumentView.corpus(corpus(object), tagger(object),
        tokenizer(object), 1, object@transforms)

      genTokDVs(object, proto)
    } else
      ref(object, "documentViews")
  })

setMethod("freqTable", "TaggedCorpusView",
  function(object, what="TaggedTokens") {
    lookup <- convertWhat(what, c("tokenTagFreqTable", "tokenFreqTable",
                                  "tagFreqTable"))

    if (is(ref(object, lookup), "nNULL")) {
      result <- NULL
      if (what == "TaggedTokens")
        result <- table(as.character(unlist.tagged(lapply(documentViews(object),
                                                          taggedTokens))))
      else if (what == "Tokens")
        result <- table(unlist(lapply(documentViews(object), tokens),
                        use.names=F))
      else
        result <- table(unlist(lapply(documentViews(object), tags),
                        use.names=F))

      if (keepComputed(object))
        (ref(object, lookup) <- result)
      else
        result
   } else
    ref(object, lookup)
 })

unlist.tagged <- function (x) {
  result <- TaggedToken()

  inner <- function (x) {
    if (is.list(x)) {
      lapply(x, inner)
    } else {
      result <<- c(result, x)
    }
  }

  inner(x)
  return(result)
}

setMethod("unique", "TaggedCorpusView",
  function (x, incomparables="missing", what="TaggedTokens") {
    lookup <- convertWhat(what, c("uniqueTokenTags", "uniqueTokens",
                                  "uniqueTags"))
    if (is(ref(x, lookup), "nNULL")) {
      if (what == "TaggedTokens")
        result <- unique(unlist.tagged(lapply(documentViews(x), unique, what=what)))
      else
        result <- unique(unlist(lapply(documentViews(x), unique, what=what)))

      if (keepComputed(x))
        (ref(x, lookup) <- result)
      else
        result
   } else
    ref(x, lookup)
 })

# See if object contains "token/tag" pair, token, or tag
setMethod("has", signature(object="TaggedCorpusView",
                           value="character"),
  function (object, value, what="TaggedTokens", ...)
    return (value %in% as.character(unique(object, what=what))))

# Inherits has for TaggedToken from TaggedView

setMethod("freq", signature(object="TaggedCorpusView",
                            value="character"),
  function (object, value, what="TaggedTokens", ...)
    if (has(object, value, what=what))
      return (freqTable(object, what=what)[[value]])
    else
      return (0))

# Inherits freq for TaggedToken from TaggedView

# Inherits documentTokenMatrix from TokenizedCorpusView

# Return the document-token/tag matrix, optionally weighted
setMethod("documentTaggedTokenMatrix", 
          signature(object="TaggedCorpusView",
                    weight="missing"),
  function (object, weight=weightTf)
    if (is(ref(object, "documentTokenTagMatrix"), "NULLTermDocMatrix")) {
      result <- makeDocumentTokenMatrix(lapply(documentViews(object),
                                           freqTable, what="TaggedTokens"),
                         as.character(unique(object, what="TaggedTokens")),
                                        names(documentViews(object)))
      if (keepComputed(object))
        (ref(object, "documentTokenTagMatrix") <- result)
      else
        result
    } else
      ref(object, "documentTokenTagMatrix"))
 
setMethod("documentTaggedTokenMatrix", 
          signature(object="TaggedCorpusView",
                    weight="WeightFunction"),
  function (object, weight=weightTf)
    weightTDM(documentTaggedTokenMatrix(object), weight))

# Return the document-tag matrix, optionally weighted
setMethod("documentTagMatrix", 
          signature(object="TaggedCorpusView",
                    weight="missing"),
  function (object, weight=weightTf)
    if (is(ref(object, "documentTagMatrix"), "NULLTermDocMatrix")) {
      result <- makeDocumentTokenMatrix(lapply(documentViews(object),
                                               freqTable, what="Tags"),
                                        unique(object, what="Tags"),
                                        names(documentViews(object)))
      if (keepComputed(object))
        (ref(object, "documentTagMatrix") <- result)
      else
        result
    } else
      ref(object, "documentTagMatrix"))
 
setMethod("documentTagMatrix", 
          signature(object="TaggedCorpusView",
                    weight="WeightFunction"),
  function (object, weight=weightTf)
    weightTDM(documentTagMatrix(object), weight))

### UTILITIES

setMethod("show", "TaggedCorpusView",
  function (object) {
    nUniqueTok <- length(unique(object, what="Tokens"))
    nUniqueTag <- length(unique(object, what="Tags"))
    nUniqueTT <- length(unique(object, what="TaggedTokens"))
    nTotalTT <- sum(documentTaggedTokenMatrix(object)@Data)

    str <- paste("A tagged corpus view of with", nTotalTT,
                 "total tagged tokens,", nUniqueTT, 
                 "unique tagged tokens,", nUniqueTok, 
                 "unique tokens, and", nUniqueTag, "unique tags\n")
    cat(str)
  }
)
