# A view of a document as a sequence of token-tag pairs

### CLASSDEF

setClass("TaggedDocumentView",
  representation("TaggedView", "TokenizedDocumentView",
    tags = reftype("character"),
    tagFreqTable = reftype("table"),
    uniqueTags = reftype("character"),
    tokenTagFreqTable = reftype("table"),
    uniqueTokenTags = reftype("TaggedToken")))

### GENERICS

setGeneric("tags", function (object) standardGeneric("tags"))

setGeneric("taggedTokens", 
  function (object) standardGeneric("taggedTokens"))

### CONSTRUCTION

# The constructor used by tag
TaggedDocumentView <- function (view, tagger, transforms=list()) {
  # TokenizedView bits
  tokens <- new("NULLcharacter")
  tokenFreqTable <- new("NULLtable")
  uniqueTokens <- new("NULLcharacter")

  # TaggedView bits
  tags <- new("NULLcharacter")
  tagFreqTable <- new("NULLtable")
  uniqueTags <- new("NULLcharacter")
  tokenTagFreqTable <- new("NULLtable")
  uniqueTokenTags <- new ("NULLTaggedToken")

  # If this is a simple tagging we can take advantage of whatever
  # computation the passed-in TokenizedDV has already done.
  if (length(transforms) == 0 && .keepComputed) { 
    tokens <- ref(view, "tokens")
    tokenFreqTable <- ref(view, "tokenFreqTable")
    uniqueTokens <- ref(view, "uniqueTokens")
  }

  # Update transforms
  transforms = c(tokenTransforms(view), transforms)

  # Copy as much as possible from the existing tokenized view
  res <- new("TaggedDocumentView",
             corpus=corpus(view),
             index=index(view),
             tokenizer=tokenizer(view),
             keepComputed=.keepComputed,
             transforms=transforms,
             tokens=tokens,
             tokenFreqTable=tokenFreqTable,
             uniqueTokens=uniqueTokens,
             tagger=tagger,
             tags=tags,
             tagFreqTable=tagFreqTable,
             uniqueTags=uniqueTags,
             tokenTagFreqTable=tokenTagFreqTable,
             uniqueTokenTags=uniqueTokenTags)

  registerView(corpus(view), res)
  return(res)
}

# A version of the constructor used by documentViews calls
TaggedDocumentView.corpus <- function (corpus, tagger, tokenizer, index,
                                       transforms=list()) {
  # TokenizedView bits
  tokens <- new("NULLcharacter")
  tokenFreqTable <- new("NULLtable")
  uniqueTokens <- new("NULLcharacter")

  # TaggedView bits
  tags <- new("NULLcharacter")
  tagFreqTable <- new("NULLtable")
  uniqueTags <- new("NULLcharacter")
  tokenTagFreqTable <- new("NULLtable")
  uniqueTokenTags <- new ("NULLTaggedToken")

  # Copy as much as possible from the existing tokenized view
  res <- new("TaggedDocumentView",
             corpus=corpus,
             index=index,
             tokenizer=tokenizer,
             keepComputed=.keepComputed,
             transforms=transforms,
             tokens=tokens,
             tokenFreqTable=tokenFreqTable,
             uniqueTokens=uniqueTokens,
             tagger=tagger,
             tags=tags,
             tagFreqTable=tagFreqTable,
             uniqueTags=uniqueTags,
             tokenTagFreqTable=tokenTagFreqTable,
             uniqueTokenTags=uniqueTokenTags)

  registerView(corpus, res)
  return(res)
}

### CALLBACK method

setMethod("clearComputed", "TaggedDocumentView",
  function (object) {
    ref(view, "tokens") <- new("NULLcharacter")
    ref(view, "tokenFreqTable") <- new("NULLtable")
    rev(view, "uniqueTokens") <- new("NULLcharacter")

    ref(object, "tags") <- new("NULLcharacter")
    ref(object, "tagFreqTable") <- new("NULLtable")
    ref(object, "uniqueTags") <- new("NULLcharacter")
    ref(object, "tokenTagFreqTable") <- new("NULLtable")
    ref(object, "uniqueTokenTags") <- new ("NULLTaggedToken")
  })

# Inherits updateView from TokenizedDocumentView
 
### ACCESSORS

# Return tokens/tags (or pairs) at given indices.  x[i, 1] returns the
# token at index i, x[i, 2] returns the tag at index i, and x[i]
# returns the token-tag pair at index i, as a TokenTag object.
setMethod("[", signature(x="TaggedDocumentView"),
  function (x, i, j, drop) {
    if (j == 1) {
      return(tokens(x)[i])
    } else if (j == 2) {
      return(tags(x)[i])
    } else stop("Second argument to extract must be 1 (token) or 2 (tag).")
  })

setMethod("[", signature(x="TaggedDocumentView", j="missing"),
  function (x, i, j, drop) TaggedToken(tokens(x)[i], tags(x)[i]))

# Little helpr method for converting what
convertWhat <- function (what, options)
  if (what == "TaggedTokens") {
    return(options[1])
  } else if (what == "Tokens") {
    return(options[2])
  } else if (what == "Tags") {
    return(options[3])
  } else stop('what must be "TaggedTokens", "Tokens", or "Tags".')

# A helper for writing funcs
# XXX think about using this/similar eventually to reduce code dup
# Needs testing
taggedMeth <- function(object, what, vars, actions) {
  lookup <- convertWhat(what, vars)

  if (is(ref(object, lookup), "nNULL")) {
    result <- NULL
    if (what == "TaggedTokens")
      result <- eval(actions[1])
    else if (what == "Tokens")
      result <- eval(actions[2])
    else
      result <- eval(actions[3])

    if (keepComputed(object))
      (ref(object, lookup) <- result)
    else
      result
  } else
    ref(object, lookup)
}

# Return a frequency table of "token/tag" pairs, tokens, or tags
setMethod("freqTable", "TaggedDocumentView",
  function (object, what="TaggedTokens") {
    # Parse what
    lookup <- convertWhat(what, c("tokenTagFreqTable", "tokenFreqTable",
                                  "tagFreqTable"))

    if (is(ref(object, lookup), "nNULL")) { # lazy compute
      # Compute the appropriate thing
      result <- NULL
      if (what == "TaggedTokens")
        result <- table(as.character(TaggedToken(tokens(object),
                                                 tags(object))))
      else if (what == "Tokens")
        result <- table(tokens(object))
      else
        result <- table(tags(object))

      # Return computed result
      if (keepComputed(object)) # save for later
        (ref(object, lookup) <- result)
      else # return and discard
        result
    } else # Load saved
      ref(object, lookup)
  })

# Returns unique "token/tag" pairs, tokens, or tags
setMethod("unique", "TaggedDocumentView", 
  function(x, incomparables="missing", what="TaggedTokens") {
    # Parse what
    lookup <- convertWhat(what, c("uniqueTokenTags", "uniqueTokens",
                                  "uniqueTags"))

    if (is(ref(x, lookup), "nNULL")) {
      result <- NULL
      if (what == "TaggedTokens") {
        result <- unique(taggedTokens(x))
       } else if (what == "Tokens") {
        result <- row.names(ref(x, "tokenFreqTable"))
        if (is.null(result))
          result <- unique(tokens(x))
       } else {
        result <- row.names(ref(x, "tagFreqTable"))
        if (is.null(result))
          result <- unique(tags(x))
       }
        
      if (keepComputed(x))
        (ref(x, lookup) <- result)
      else
        result
    } else
      ref(x, lookup)
  })

# See if object contains "token/tag" pair, token, or tag
setMethod("has", signature(object="TaggedDocumentView",
                           value="character"),
  function (object, value, what="TaggedTokens", ...)
    if (what == "TaggedTokens")
      return(value %in% as.character(taggedTokens(object)))
    else if (what == "Tokens")
      return (value %in% tokens(object))
    else if (what == "Tags")
      return (value %in% tags(object))
    else stop('what must be "TaggedTokens", "Tokens", or "Tags".'))

# inherits has for TaggedToken from TaggedView

# Return the frequency of the given "token/tag" pair, token, or tag
setMethod("freq", signature(object="TaggedDocumentView",
                            value="character"),
  function (object, value, what="TaggedTokens", ...)
    if (has(object, value, what=what))
      return (freqTable(object, what=what)[[value]])
    else
      return (0))

# inherits freq for TaggedToken from TaggedView

# Returns the tags of this view
setMethod("tags", "TaggedDocumentView",
  function (object)
    if (is(ref(object, "tags"), "NULLcharacter"))
      if (keepComputed(object))
        (ref(object, "tags") <-
          tag(tokens(object), tagger(object)))
      else
        tag(tokens(object), tagger(object))
    else
      ref(object, "tags"))

# Return token-tag pairs in "token/tag" format
setMethod("taggedTokens", "TaggedDocumentView",
  function (object)
    TaggedToken(tokens(object), tags(object)))

### UTILITIES
setMethod("show", "TaggedDocumentView",
  function (object) {
    name <- ID(document(object))
    nTTs <- length(taggedTokens(object))
    nUniqueTokens <- length(unique(object, what="Tokens"))
    nUniqueTags <- length(unique(object, what="Tags"))
    nUniqueTTs <- length(unique(object))

    str <- paste("A tagged document view of", name, "with", nTTs, 
                 "total tagged tokens,", nUniqueTTs,
                 "unique tagged tokens,", nUniqueTokens, 
                 "unique tokens, and", nUniqueTags, "unique tags\n")
    cat(str)
  }
)
