# A simple pair object representing a tagged token, or vector of
# tag-token pairs.  Tags can currently only be strings, but we may
# eventually allow any type.

### CLASSDEF

setClass("TaggedToken",
  representation(token = "character", tag = "character"))

### GENERICS

setGeneric("tag",
  function (object, tagger, tokenizer, index, ...) standardGeneric("tag"))
setGeneric("token", function (object) standardGeneric("token"))
setGeneric("token<-", function (object, value) standardGeneric("token<-"))

# Tagger provides tag generic...if weirdly
setGeneric("tag<-", function (object, value) standardGeneric("tag<-"))

### CONSTRUCTION

TaggedToken <- function (token=character(), tag=character()) {
  if (!is.character(token))
    stop(paste("Bad token: ", token))
  if (!is.character(tag))
    stop(paste("Bad tag: ", tag))
  if (length(token) != length(tag))
    stop("Token and tag vectors do not conform")

  new("TaggedToken", token=token, tag=tag)
}

### ACCESSORS

# Return the tagged token as a token/tag string
setMethod("as.character", "TaggedToken",
  function (x) paste(x@token, x@tag, sep="/"))

# Return the tagged token as an alternating character vector
setMethod("as.vector", signature("TaggedToken", "missing"),
  function (x, mode="any") c(rbind(x@token, x@tag)))

# Get the token
setMethod("token", "TaggedToken",
  function (object) object@token)

# Get the tag
setMethod("tag", signature("TaggedToken", "missing", "missing", "missing"),
  function (object) object@tag)

setMethod("[", signature(x = "TaggedToken", j = "missing", drop = "missing"),
  function (x, i, j, drop) TaggedToken(x@token[i], x@tag[i]))

setMethod("[", signature(x = "TaggedToken", j = "numeric", drop = "missing"),
  function (x, i, j, drop) 
    if (j == 1) token(x)[i]
    else if (j == 2) tag(x)[i]
    else stop(paste("Index j =", j, "must be 1 or 2")))

# Some important generics

# XXX why doesn't %in% work?
setGeneric("match")
setMethod("match", signature(x="TaggedToken", table="TaggedToken"),
  function (x, table, nomatch = NA_integer_, incomparables = NULL)
    match(as.character(x), as.character(table), nomatch, incomparables))
 
setGeneric("unique")
setMethod("unique", "TaggedToken",
  function (x, incomparables="missing")
    x[!duplicated(as.character(x))])


c2TT <- function (x, y) new("TaggedToken", token=c(token(x), token(y)),
                                           tag=c(tag(x), tag(y)))

# Cat to TaggedToken object
setMethod("c", signature(x = "TaggedToken"),
  function (x, ..., recursive=FALSE) {
    args <- list(...)
    res <- x
    for (arg in args)
      res <- c2TT(res, arg)
    res
  })

setMethod("length", "TaggedToken",
  function (x) length(token(x)))

### MODIFIERS

# Set the token
setMethod("token<-", "TaggedToken",
  function (object, value) {
    if (!is.character(value))
      stop(paste("Bad token: ", value))
    if (length(value) != length(object@token))
      stop("Token vector is wrong length")
    object@token <- value
    object
  })

# Set the tag
setMethod("tag<-", "TaggedToken",
  function (object, value) {
    if (!is.character(value))
      stop(paste("Bad tag: ", value))
    if (length(value) != length(object@tag))
      stop("Tag vector is wrong length")
    object@tag <- value
    object
  })

setMethod("[<-", signature(x = "TaggedToken", j = "numeric", value="character"),
  function (x, i, j, ..., value) {
    if (j == 1) x@token[i] <- value
    else if (j == 2) tag(x)[i] <- value
    else stop(paste("Index j =", j, "must be 1 or 2"))
    x
  })

### SHOW

  # Print as token/tag
  setMethod("show", "TaggedToken",
    function (object)
      print(as.character(object)))
