# A base class for all classes implementing the tagger interface.  

# Tagger defines two methods: tag, and batch.tag.  

# tag takes an object representing a sequence of tokens (or if given
# an untokenized object, takes that object and applies the default
# tokenizer to it) and applies a tag to each token.  Extending class
# must provide a method definition of the form: 
#
# setMethod("tag", signature(object="character", tagger="MyTagger"),  ...  
#
# They should use the helper function at the bottom of this file to
# tokenize non-tokenized character strings.  The methods below take
# care of breaking down higher-order types into underlying character
# data.

### CLASS REF
### Defined, as below, in TaggedView.R for technical reasons.
#setClass("Tagger", representation("VIRTUAL"))

### GENERICS

# Below is in TaggedToken.R for load order
#setGeneric("tag",
#  function (object, tagger, tokenizer, index, ...) standardGeneric("tag"))
setGeneric("batch.tag", 
  function (object, tagger) standardGeneric("batch.tag"))

tag.character.helper <- function(object, tokenizer) {
  tokens <- NULL
  if (length(object) == 1)
    tokens <- tokenize(object, tokenizer)
  else
    tokens <- object
  return(tokens)
}

### TAG OVERLOADINGS

# The default tag behavior for characters is to assume that the
# character object contains a sequence of tokens when no tokenizer is
# given, but to tokenize the object first, if a tokenizer is
# specifically provided.  Note that we explicitly override the method
# for PlainTextDocument objects to provide tokenization in all cases.

# The method signature all extenders must implement
setMethod("tag", signature(object="character", tagger="Tagger",
                           tokenizer="missing", index="missing"),
  function (object, tagger, tokenizer, index)
    stop("tag must be implemented by derived class"))

# Calls the above method when a tokenizer is provided, after
# tokenizing
setMethod("tag", signature(object="character", tagger="Tagger",
                           tokenizer="Tokenizer", index="missing"),
  function (object, tagger, tokenizer, index)
      tag(tokenize(object, tokenizer), tagger))

# Now for non-character types

# Generate a TaggedDocumentView from a TokenizedDocumentView.
setMethod("tag", signature(object="TokenizedDocumentView",
                           tagger="Tagger",
                           tokenizer="missing",
                           index="missing"),
  function (object, tagger, tokenizer, index)
    TaggedDocumentView(object, tagger))

# Generate a TaggedCorpusView from a TokenizedCorpusView
setMethod("tag", signature(object="TokenizedCorpusView",
                           tagger="Tagger",
                           tokenizer="missing",
                           index="missing"),
  function (object, tagger, tokenizer, index)
    TaggedCorpusView(object, tagger))

# Generate a TaggedDocumentView from a TokenizedCorpusView + index
setMethod("tag", signature(object="TokenizedCorpusView",
                           tagger="Tagger",
                           tokenizer="missing",
                           index="numeric"),
  function (object, tagger, tokenizer, index)
    TaggedDocumentView(object[[index]], tagger))

# Generate a TaggedCorpus from a PCorpus.
setMethod("tag", signature(object="PCorpus", 
                           tagger="Tagger",
                           index="missing"),
  function(object, tagger, tokenizer, index) 
    if (missing(tokenizer))
      tag(tokenize(object), tagger)
    else
      tag(tokenize(object, tokenizer), tagger))

# Generate a TaggedDocumentView from a PCorpus + index.
setMethod("tag", signature(object="PCorpus",
                           tagger="Tagger",
                           index="numeric"),
  function (object, tagger, tokenizer, index)
    if (missing(tokenizer))
      tag(tokenize(object), tagger, index=index)
    else
      tag(tokenize(object, tokenizer), tagger, index=index))

# batch.tag takes an object representing a set of sentences and tags
# each sentence.  By default, this object should be a list, with one
# "sentence" per list item.  Each sentence is of the form expected by
# tag.  Extending classes need only override batch.tag if they require
# behavior other than the default of applying tags to each sentence
# and returning a list of tagged sentences, which the following
# definitions take care of.

# Apply the tagger to a list of sentences.  
setMethod("batch.tag", signature(object="list",
                                 tagger="Tagger"),
  function(object, tagger) lapply(object, tag))

# TODO batch_tag for whatever type we use to represent documents that
# are broken up into sentences/parts.


# Helper function for taggers that train on a set of tagged sentences,
# represented in nested list form

"goodTaggedSentenceList" <-
# Function to make sure a training set in nested list form is
# formatted correctly
function(tslist) {
  lapply(tslist, function (sublist) {
    if (! is.list(sublist))
      return(FALSE)
    lapply(sublist, function(pairlist) {
      if (! is.list(pairlist))
        return(FALSE)
      if (length(pairlist) != 2)
        return(FALSE)
      if (! is.character(pairlist[[1]]))
        return(FALSE)
      if (! is.character(pairlist[[2]]))
        return(FALSE)
     })
   })

   return(TRUE)
}
