\name{MatchObject-class}
\docType{class}
\alias{MatchObject-class}
\alias{MatchObject-methods}
\alias{MatchObject}
\alias{is.na}
\alias{re.endpos}
\alias{re.end}
\alias{re.expand}
\alias{re.group}
\alias{re.groups}
\alias{re.lastgroup}
\alias{re.lastindex}
\alias{re.pos}
\alias{re.re}
\alias{re.span}
\alias{re.start}
\alias{re.string}
\alias{show,MatchObject-method}

\title{Class "MatchObject"}

\description{A MatchObject represents the result of a regular
expression match returned by the \code{match} and \code{search}
methods of the \code{\linkS4class{RegexObject}}.}

\section{Objects from the Class}{
Users should not create MatchObjects directly.
}

\section{Methods}{
  \describe{
    \item{re.expand}{\code{signature(match = "MatchObject", template = "character")}: Performs backslash substitution on the string \code{template} and returns the result.  Both standard escapes, such as \code{"\\n"} and numeric and named backreferences (e.g.  \code{"\\1"}, \code{"\\g<1>"}, and \code{"\\g<name>"}) are processed and replaced with appropriate text.}
    \item{re.group}{\code{signature(match = "MatchObject")}: Returns one or more subgroups of a match.  By default the method returns a single-element \code{character} vector containing the entire match string.  Callers may optionally pass one or more group indicators (these are single integers or strings), in which case the method returns a vector of matches, with one match for each requested subgroup.  The entire match is represented by the numeric indicator 0.  An error occurrs if a requested subgroup does not exist.  The method returns an empty string for any requested subgroup that did not take part in the match generating \code{match}.}
    \item{re.groups}{\code{signature(match = "MatchObject")}: Returns a list containing all the subgroups of the match.  Subgroups that did not participate in the match are represented by \code{NA}, although the caller may pass an optional argument to take the place of this default represenation.}
    \item{re.start}{\code{signature(match = "MatchObject")}: Returns the index of the character at the start of substring matched by the optional \code{group} argument.  \code{group} defaults to 0, indicating the entire match, but can be set to any existing group.  The method returns -1 if \code{group} exists but did not contribute to the match, and generates an error if \code{group} does not exist.}
    \item{re.end}{\code{signature(match = "MatchObject")}: Returns the index of the character at the end of substring matched by the optional \code{group} argument.  \code{group} defaults to 0, indicating the entire match, but can be set to any existing group.  The method returns -1 if \code{group} exists but did not contribute to the match, and generates an error if \code{group} does not exist.}
    \item{re.span}{\code{signature(match = "MatchObject")}: Returns a vector containing the indices of the characters at the start and end of substring matched by the optional \code{group} argument.  \code{group} defaults to 0, indicating the entire match, but can be set to any existing group.  The method returns \code{(-1, -1)} if \code{group} exists but did not contribute to the match, and generates an error if \code{group} does not exist.}
    \item{re.pos}{\code{signature(match = "MatchObject")}: The value of \code{pos} that was passed to the \code{search} or \code{match} method of \code{RegexObject} when generating \code{match}.}
    \item{re.endpos}{\code{signature(match = "MatchObject")}: The value of \code{endpos} that was passed to the \code{search} or \code{match} method of \code{RegexObject} when generating \code{match}.}
    \item{re.lastindex}{\code{signature(match = "MatchObject")}:  The index of the last matching capture group used in the \code{\linkS4class{RegexObject}} that generated \code{match}, or \code{NULL} if no group matched at all.}
    \item{re.lastgroup}{\code{signature(match = "MatchObject")}: The name of the last matching capture group used in the \code{\linkS4class{RegexObject}} that generated \code{match}, or \code{NULL} if the group didn't have a name, or no group matched at all.}
    \item{re.re}{\code{signature(match = "MatchObject")}: Returns the \code{\linkS4class{RegexObject}} whoe \code{match} or \code{search} method produced \code{match}.}
    \item{re.string}{\code{signature(match = "MatchObject")}: Returns the string passed to \code{match} or \code{search} to generate \code{match}.}
	 }
}

\note{All of the above methods also dispatch on objects of type
\code{list} in which case the above methods are applied to every
object in the list.}

\references{Python Library Reference: re -- Regular expression
operations. http://www.python.org/doc/lib/module-re.html.}

\author{Dan Pemstein <dpemstein@iq.harvard.edu>}

\seealso{
  \code{\linkS4class{RegexObject}}
}

\examples{
# Some strings to match.  Note that we can apply RegexObject methods
# directly to the string vector and they will return vectors/lists of
# results.
str <- c("123", "??", "123xxx", "12?af?123fd", "!!!", "Hello, how are you?")

# Matching - notice the double-escaping
p1 <- re.compile('(\\\d+)|([a-z]+)+', ignorecase=TRUE)
m1 <- re.match(p1, str)

# Searching - again, notice the double-escaping
p2 <- re.compile('([\\\?\\\!]+)|(h[a-z]+)')
m2 <- re.search(p2, str)

# Play around with group 
g1 <- re.group(m1)
g2 <- re.group(m1, 0, 1, 2)
g3 <- re.group(m2, 0, 1, 2)

# More group, now with named subgroups
p5 <- re.compile('(?P<bigh>H[a-z]+)(\\\W+)(?P<littleh>h[a-z]+)')
m3 <- re.search(p5, str)
g4 <- re.group(m3, 0, 1, 2, 3, "bigh", "littleh")

# Now groups
gs1 <- re.groups(m2)
gs2 <- re.groups(m2, "DOH")

# Get match start indices
st1 <- re.start(m2)
st2 <- re.start(m2, 1)
st3 <- re.start(m2, 2)

# Get match end indices
en1 <- re.end(m2)
en2 <- re.end(m2, 1)
en3 <- re.end(m2, 2)

# Get match spans
sp1 <- re.span(m2)
sp2 <- re.span(m2, 1)
sp3 <- re.span(m2, 2)
}

\keyword{classes}
