#ifndef __EMBED_NLTK__
#define __EMBED_NLTK__

/* This file contains basic routines for starting up and shutting down
 * the embedded python interpreter and importing the needed nltk
 * modules, and also a number of utility functions for converting
 * between R and python objects.
 *
 * Only init_python and finalize_python methods are designed to be
 * called directly from R.
 */

#include <Python.h>
#include <Rinternals.h>
#include <stdbool.h>
#include <stdarg.h>

/* Globals - only accessible from this file */
static PyObject* PYTHON_MAIN;        /* Pointer to __main__ module */
static PyObject* PYTHON_RE;          /* Pointer to re module */
static PyObject* NLTK_TAG;           /* Pointer to nltk tag module */
static PyObject* NLTK_PUNKT;         /* Pointer to nltk punkt module */

/* An imp-indep way to get a pointer to the __main__ module */
inline PyObject*
python_main_pnt()
{
  return PYTHON_MAIN;
}

/* An imp-indep way to get a pointer to the re module */
inline PyObject*
python_re_pnt()
{
  return PYTHON_RE;
}

/* An imp-indep way to get a pointer to the nltk.tag module */
inline PyObject*
nltk_tag_pnt ()
{
  return NLTK_TAG;
}

/* An imp-indep way to get pointer to the nltk.tokenize.punkt module */
inline PyObject*
nltk_punkt_pnt ()
{
  return NLTK_PUNKT;
}

/* Start up the embedded python interpreter and import the necessary
 * nltk modules.  
 *
 * Designed to be called from R with .Call.  Returns nil on exit.
 */
inline SEXP
init_python ()
{
  Py_Initialize();
  PYTHON_MAIN = PyImport_AddModule("__main__");
  PYTHON_RE = PyImport_ImportModule("re");
  NLTK_TAG = PyImport_ImportModule("nltk.tag");
  NLTK_PUNKT = PyImport_ImportModule("nltk.tokenize.punkt");

  return R_NilValue;
}

/* Decref imported modules and close down the embedded interpreter.
 *
 * Designed to be called from R with .Call.  Returns nil on exit.
 */
inline SEXP
finalize_python ()
{
  Py_DECREF(python_re_pnt());
  Py_DECREF(nltk_tag_pnt());
  Py_DECREF(nltk_punkt_pnt());
  Py_Finalize();

  return R_NilValue;
}

/* Construct a python object given a module, argument tuple, and
 * object name, with some error checking.
 *
 * The caller is responsible for Py_DECREFing the returned object.
 *
 * !!! Note that this function steals the argtuple reference (that is,
 * it decrefs it before returning) but not the module reference, which
 * it borrows.
 */
inline PyObject*
construct_python_object (PyObject* module, PyObject* argtuple,
                         const char* name)
{ 
  PyObject *constructor, *object;

  /* Grab the constructor */
  constructor = PyObject_GetAttrString(module, name);
  if (constructor == NULL)  {
    PyErr_Print();
    Rprintf("%s: ", name);
    error("Error fetching python object constructor");
  }

  /* Do the construction */
  object = PyObject_CallObject(constructor, argtuple);
  if (object == NULL) {
    PyErr_Print();
    Rprintf("%s: ", name);
    error("Error constructing python object");
  }

  /* Clean up a bit */
  Py_DECREF(argtuple); /* dispose of used argument */
  Py_DECREF(constructor); /* dispose of the constructor ref */


  return object;
}

/* Wraps an existing python object in an R external pointer, and
 * returns the points.  Takes a Pyobject pointer and a destructor
 * functio to call when the python object is no longer referenced by
 * any R objects.
 *
 * !!! Note the caller is responsible for unprotecting the returned
 * pointer before retruning to R.  Furthermore, the caller is
 * responsible for Py_DECREFing the wrapped python object; normally,
 * this is done by the passed-in destructor function.
 */
SEXP
wrap_python_object(PyObject* obj, R_CFinalizer_t destructor)
{
  SEXP ret;

  /* Create the external pointer and register the destructor */
  PROTECT(ret = R_MakeExternalPtr(obj, R_NilValue, R_NilValue));
  R_RegisterCFinalizerEx(ret, destructor, TRUE);
  
  return ret;
}

/* Constructs a Python object and returns an R external pointer to it.
 * Takes a python module to find the object constructor in and a tuple
 * containing arguments to the constructor, the name of the
 * constructor as a string, and a destructor function to call when the
 * python object is no longer referenced by any R objects.  Returns an
 * SEXP pointing to an extenalptr object (which points to the python
 * object).
 *
 * !!! Note that this function steals the argtuple reference (that is,
 * it decrefs it before returning) but not the module reference, which
 * it borrows.  Also, the caller is responsible for unprotecting the
 * returned pointer before returning to R.  Furthermore, the caller is
 * responsible for Py_DECREFing the wrapped python object.  Normally,
 * this is done by the passed-in destructor function.
 */
SEXP
construct_wrapped_python_object(PyObject* module, PyObject* argtuple, 
                                const char* name, R_CFinalizer_t destructor)
{
  PyObject *object = construct_python_object(module, argtuple, name);
  return wrap_python_object(object, destructor);
}

/* Clean up a python object, given its R external pointer */
inline void
destruct_python_object (SEXP externalptr)
{
  PyObject* object = (PyObject*) R_ExternalPtrAddr(externalptr);
  Py_DECREF(object);
}


/**** Utility functions (mostly R-c-python data type converters) ****/

/* Call a python method given a python object to invoke on, a method
 * name, and a tuple containing a arguments.  Also, takes a boolean
 * argument indicating whether or not to Py_DECREF the argument tuple
 * before returning.
 *
 * The caller is responsibel for Py_DECREFing the returned python
 * object.
 *
 */

inline PyObject*
call_python_method (PyObject* object, char* methname,
                    PyObject* argtuple, bool decref_arg)
{
  PyObject *retval, *ptype, *pvalue, *ptraceback, *errtype, *errval, *meth;
  if (argtuple == NULL || PyTuple_Size(argtuple) == 0) {
    meth = PyString_FromString(methname);
    retval = PyObject_CallMethodObjArgs(object, meth, NULL);
   } else
    retval = PyObject_CallMethod(object, methname, "N", argtuple);
  if (PyErr_Occurred() != NULL) {
    PyErr_Fetch(&ptype, &pvalue, &ptraceback);
    errtype = PyObject_Str(ptype);
    errval = PyObject_Str(pvalue);
    if (ptype != NULL)
      Py_DECREF(ptype);
    if (pvalue != NULL)
      Py_DECREF(pvalue);
    if (ptraceback != NULL)
      Py_DECREF(ptraceback);
    error("%s: %s\n", PyString_AsString(errtype),
            PyString_AsString(errval));
    Py_DECREF(errtype);
    Py_DECREF(errval);
  }
  if (decref_arg && argtuple != NULL)
    Py_DECREF(argtuple);
  return retval;
}

/* Like the above method but takes a variable number of python objects
 * instead of a tuple.  STEALS references to the passed-in objects.
 */

inline PyObject*
call_python_method_va (PyObject* object, char* methname, int n_args, ...)
{
  va_list argp;
  PyObject *retval, *arg, *args;

  va_start(argp, n_args);
  args = PyTuple_New(n_args);
  for (int i = 0; i < n_args; ++i) {
    arg = va_arg(argp, PyObject*);
    PyTuple_SetItem(args, i, arg); /* Steals ref */
  }
  va_end(argp);

  retval = call_python_method(object, methname, args, true);

  return retval;
}

/* Extract a python object from an external pointer SEXP.
 *
 * TODO perhaps more error checking
 */
inline PyObject*
get_pyobj_from_extptr (SEXP ptr)
{
  PyObject* obj = (PyObject*) R_ExternalPtrAddr(ptr);
  if (obj == NULL)
    error("Wrapped python object address is invalid");
  return obj;
}

/* Extract an attribute from a python object pointed to by an external
 * pointer.
 *
 * !!! The caller is responsible for Py_DECREFing the returned
 * object.
 *
 * TODO perhaps more error checking
 */
inline PyObject*
get_pyattr_from_extptr (SEXP ptr, const char *attr_name)
{
  PyObject *obj, *attr;
  obj = get_pyobj_from_extptr(ptr);
  attr = PyObject_GetAttrString(obj, attr_name);
  return attr;
}


/* A generalization of the CHAR macro in Rdefines.h.  This macro
 * returns a pointer to an array of pointers to character arrays,
 * where each pointer points to a string in the character vector x.
 *
 * Basically, you can get a pointer to the underlying c string in a
 * CHARSXP but CHAR, unlike integer, doesn't return an array of
 * pointers for a STRSXP.  We want this to avoid copying a bunch of
 * strings, although we still need to allocate a pointer array.
 *
 * !!! The one tricky thing is that you need to free your array.
 *
 * XXX Leaving CHARVEC code in for now just in case, but not looking
 * like we're going to need it.
 */
#define CHARVEC(x) point_to_chars(x)

inline const char**
point_to_chars(SEXP stringvec)
{
  int size = LENGTH(stringvec);

  const char** array =
    (const char**) malloc (sizeof(char**[size]));

  for (int i = 0; i < size; ++i)
    array[i] = CHAR(STRING_ELT(stringvec, i));

  return array;
}

/* A helper that converts the python object to the python string "" if
 * if it python's None, and just lets the string pass through
 * otherwise.
 */

inline PyObject*
convert_none_str(PyObject* string)
{
  if (PyString_Check(string))
    return string;
  else if (string == Py_None)
    return PyString_FromString("");
  else
    error("Attempted de-none-ify non-string");
}


/* Takes a 1-element R character vector and returns a python string.
 *
 * The caller is responsible for Py_DECREFing the returned object.
 */
inline PyObject*
r2p_string (SEXP r_string)
{
  PyObject* string = PyString_FromString(CHAR(STRING_ELT(r_string, 0)));
  return string;
}

/* Takes a python string and returns an R character vector (length 1)
 * containing the same string.
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_string (PyObject* p_string)
{
  SEXP string;
  PROTECT(string = allocVector(STRSXP, 1));
  SET_STRING_ELT(string, 0, mkChar(PyString_AsString(p_string)));
  return string;
}

/* Takes a 1-element R integer vector and returns a python string.
 *
 * The caller is responsible for Py_DECREFing the returned object.
 */
inline PyObject*
r2p_int (SEXP r_int)
{
  PyObject* integer = PyInt_FromLong((long) INTEGER(r_int)[0]);
  return integer;
}

/* Takes a python integer and returns a 1-elem R int vector
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_int (PyObject* p_int)
{
  SEXP integer;
  PROTECT(integer = allocVector(INTSXP, 1));
  INTEGER(integer)[0] = (int) PyInt_AsLong(p_int);
  if (PyErr_Occurred() != NULL)
    error("Tried to convert a non-integer in p2r_int");
  return integer;
}

/* Takes an R string vector of tokens and returns a pointer to a
 * python list containing the same tokens.
 *
 * The caller is responsible for Py_DECREFinf the returned object.
 */
inline PyObject*
r2p_token_list(SEXP r_list)
{
  PyObject *list;
  int length = LENGTH(r_list);
  
  list = PyList_New(length);
  for (int i = 0; i < length; ++i)
    PyList_SetItem(list, i, 
        PyString_FromString(CHAR(STRING_ELT(r_list, i))));

  return list;
}

/* More general alias for above */
#define r2p_string_list(x) r2p_token_list(x)

/* Takes a python list of tokens and returns an R string vector
 * containing the same tokens.
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_token_list(PyObject *p_list)
{
  SEXP list;
  int length = PyList_Size(p_list);

  PROTECT(list = allocVector(STRSXP, length));
  for (int i = 0; i < length; ++i) {
    SET_STRING_ELT(list, i, 
        mkChar(PyString_AsString(PyList_GetItem(p_list, i))));
  }

  return list;
}

/* More general alias for above */
#define p2r_string_list(x) p2r_token_list(x)

/* Takes a python list containing either a string, or a tuple of
 * strings in each slot and returns a list of string vectors.
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_stringorstringtuple_list(PyObject *p_list)
{
  SEXP list, cvec;
  int length, tlength;
  PyObject *tmp;

  length = PyList_Size(p_list);
  PROTECT(list = allocVector(VECSXP, length));
  for (int i = 0; i < length; ++i) {
    tmp = PyList_GetItem(p_list, i);
    if (PyTuple_Check(tmp)) { // if tuple
      tlength = PyTuple_Size(tmp);
      PROTECT(cvec = allocVector(STRSXP, tlength));
      for (int j = 0; j < tlength; ++j)
        SET_STRING_ELT(cvec, j, 
            mkChar(PyString_AsString(PyTuple_GetItem(tmp, j))));
    } else if (PyString_Check(tmp)) {
      PROTECT(cvec = allocVector(STRSXP, 1));
      SET_STRING_ELT(cvec, 0, 
          mkChar(PyString_AsString(tmp)));
    } else { // We're checking anyway, so be safer here than usual
      error("Passed non-compatible object to p2r_stringostringtuple_list");
    }
    SET_VECTOR_ELT(list, i, cvec);
    UNPROTECT(1); // cvec
  }

  return list;
}
/* Takes a python string tuple and returns an R character vector.
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_string_tuple (PyObject* p_tuple)
{
  SEXP list;
  int length;
 
  if (! PyTuple_Check(p_tuple)) {
    PROTECT(list = allocVector(STRSXP, 1));
    SET_STRING_ELT(list, 0, 
        mkChar(PyString_AsString(convert_none_str(p_tuple))));
  } else {
    length = PyTuple_Size(p_tuple);
    PROTECT(list = allocVector(STRSXP, length));
    for (int i = 0; i < length; ++i) {
      SET_STRING_ELT(list, i, 
        mkChar(PyString_AsString(
            convert_none_str(PyTuple_GetItem(p_tuple, i)))));
    }
  }

  return list;
}

/* Takes an R list representing a tagged string---that is a list of
 * lists where each sublist is a pair containing two string, a token
 * and a tag---and returns a pointer to the equivalent python object
 * (which is a list of tuple pairs).
 *
 * The caller is responsible for Py_DECREFinf the returned object.
 */
inline PyObject*
r2p_tagged_token_list (SEXP r_list)
{
  SEXP r_tokentag;
  PyObject *list, *tokentag;
  int size;
 
  size = LENGTH(r_list);
  list = PyList_New(size);
  for (int i = 0; i < size; ++i) {
    r_tokentag = VECTOR_ELT(r_list, i);
    tokentag = PyTuple_New(2);
    PyTuple_SetItem(tokentag, 0, 
      PyString_FromString(CHAR(STRING_ELT(VECTOR_ELT(r_tokentag, 0), 
            0))));
    PyTuple_SetItem(tokentag, 1, 
      PyString_FromString(CHAR(STRING_ELT(VECTOR_ELT(r_tokentag, 1), 
            0))));
    PyList_SetItem(list, i, tokentag);
  }

  return list;
}

/* Takes a python list of tagged tokens (each in a tuple pair) and
 * returns an R list
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_tagged_token_list(PyObject *p_list)
{
  PyObject *p_tokentag;
  SEXP list, tokentag;
  int length = PyList_Size(p_list);
  
  PROTECT(list = allocVector(VECSXP, length));
  for (int i = 0; i < length; ++i) {
    p_tokentag = PyList_GetItem(p_list, i);
    PROTECT(tokentag = allocVector(STRSXP, 2));
    SET_STRING_ELT(tokentag, 0, 
        mkChar(PyString_AsString(convert_none_str(PyTuple_GetItem(p_tokentag, 0)))));
    SET_STRING_ELT(tokentag, 1, 
        mkChar(PyString_AsString(convert_none_str(PyTuple_GetItem(p_tokentag, 1)))));
    SET_VECTOR_ELT(list, i, tokentag);
    UNPROTECT(1); /* tokentag is now protected by list */
  }

  return list;
}

/* Takes a python list of tagged tokens (each in a tuple pair) and
 * returns a 2-columnt R matrix.
 *
 * The caller is responsible for UNPROTECTing the returned object.
 */
inline SEXP
p2r_tagged_token_list2matrix (PyObject* p_list)
{
  PyObject *p_tokentag;
  SEXP matrix, tokentag;
  int length = PyList_Size(p_list);
  
  PROTECT(matrix = allocMatrix(STRSXP, length, 2));
  for (int i = 0; i < length; ++i) {
    p_tokentag = PyList_GetItem(p_list, i);
    SET_STRING_ELT(matrix, i, 
        mkChar(PyString_AsString(convert_none_str(PyTuple_GetItem(p_tokentag, 0)))));
    SET_STRING_ELT(matrix, i + length, 
        mkChar(PyString_AsString(convert_none_str(PyTuple_GetItem(p_tokentag, 1)))));
  }

  return matrix;
}

/* Takes an R list representing a set of tagged sentences---that is, a
 * list of lists where each sublist represents a sentence and contains
 * a series of lists (pairs) each containing a string tag and a
 * token---and returns a pointer to the equivalent python object
 * (which is a list of lists of tuple pairs).
 *
 * The caller is responsible for Py_DECREFing the returned object.
 */
inline PyObject*
r2p_tagged_token_sentences (SEXP r_list)
{
  PyObject *list;
  int nsent = LENGTH(r_list);

  list = PyList_New(nsent);
  for (int i = 0; i < nsent; ++i)
    PyList_SetItem(list, i, 
        r2p_tagged_token_list(VECTOR_ELT(r_list, i)));

  return list;
}

/* Takes an R list representing a set of tokenized sentences---that
 * is, a list of character vectors where each character vector
 * represents a tokenized sentence, one token per vector element---and
 * returns a point to a the equivalent python object (which is a list
 * of lists).
 *
 * The caller is responsible for Py_DECREFing the returned object.
 */
inline PyObject*
r2p_token_sentences (SEXP r_list)
{
  PyObject *list;
  int nsent = LENGTH(r_list);

  list = PyList_New(nsent);
  for (int i = 0; i < nsent; ++i)
    PyList_SetItem(list, i, r2p_token_list(VECTOR_ELT(r_list, i)));

  return list;
}

/* Various R-conversions for special cases of the
 * get_pyattr_from_extptr.  Return NULL if attribute == Py_None.
 *
 * !!! Caller is responsible for UNPROTECTing the returned object
 */

inline SEXP
get_intattr_from_extptr (SEXP ptr, const char *attr_name)
{
  PyObject *attr;
  SEXP r_attr;

  attr = get_pyattr_from_extptr(ptr, attr_name);
  if (attr == Py_None)
    r_attr = R_NilValue;
  else
    r_attr = p2r_int(attr);
  Py_DECREF(attr);

  return r_attr;
}

inline SEXP
get_strattr_from_extptr (SEXP ptr, const char *attr_name)
{
  PyObject *attr;
  SEXP r_attr;

  attr = get_pyattr_from_extptr(ptr, attr_name);
  if (attr == Py_None)
    r_attr = R_NilValue;
  else
    r_attr = p2r_string(attr);
  Py_DECREF(attr);

  return r_attr;
}

#endif /* __EMBED_NLTK__ */
