#ifndef __LOOKUP_TABLE__
#define __LOOKUP_TABLE__

/* This file contains the necessary functions to create and manage a
 * lookup-table for nltk python objects.
 *
 * XXX This header is no longer in use because Dan discovered external
 * pointer objects.  For now, remains in the repository for possible
 * future reference.  Note that this requires glib2 for the hash.
 */

#include <Python.h>
#include <limits.h>
#include <glib.h>
#include <stdio.h>
#include <stdbool.h>

/* We need to keep our own address table because R integers are always
 * expressed in terms of ints, which can't hold addresses on certain
 * 64 bit platforms.  So we keep a lookup table with ints as keys and
 * PyObject pointers as values.  This probably means we can't use all
 * of the address space available to us on many systems, but this
 * seems unlikely to be an issue since the average user shouldn't need
 * to hold too many nltk objects at once.  These functions control
 * initialization, access, modification, and destruction of the lookup
 * table.  We use ints as indexes because that is all we can pass to R
 * (even though the table could be larger in theory).  Indices start
 * at INT_MIN and go up to INT_MAX and are reused as they become
 * available.
 *
 * At the moment this is implemented in terms of glib's GHashTable
 * objects, but we could create our own hash implementation to avoid
 * the dependency, if need be.  Index reuse is facilitated by a free
 * index stack.  The free stack uses a simple squaring-halving resize
 * scheme.  Whenever a free index resides on the stack it is used for
 * the next object allocation, otherwise a new index is generated by
 * incrementing the current index integer.  Therefore, even if the
 * user rapidly creates and destroys a bunch of nltk objects, we're
 * unlikely to run out of indices with this scheme, at least before
 * the machine runs out of memory.
 */

/* The lookup table struct.  Consists of a hash, a current index
 * counter, and a free value stack for reusing freed indices.
 */
struct lookup_table {
  int* free_stack;
  int free_stack_end;  /* XXX these two should be long or whatever */
  int free_stack_size; /* signed type can index (INT_MIN, INT_MAX) */
  int current_index;
  GHashTable* table;
};

/* The base size of the free index stack.  This is the max number of
 * nltk objects the user can create without causing a stack resize.
 * Note that this number should never be smaller than 2.
 */
#define FREE_STACK_INIT_SIZE 32

/* Checks if the free stack in the given lookup table is empty */
inline bool
is_free_stack_empty (struct lookup_table* lookup)
{
  if ((*lookup).free_stack_end > -1)
    return false;
  return true;
}

/* Pops and returns the value on the top of the free stack for the
 * given lookup table.  Resizes the memory allocated for the free
 * stack as necessary.  Behavior is undefined if the free stack is
 * empty.
 */
inline int
get_free_stack_val (struct lookup_table* lookup)
{
  int retval, **free_stack, *free_stack_end, *free_stack_size;
 
  /* Some shorthand */
  free_stack = &((*lookup).free_stack);
  free_stack_end = &((*lookup).free_stack_end);
  free_stack_size = &((*lookup).free_stack_size);

  /* Get the index value, and shrink the stack */
  retval = (*free_stack)[(*free_stack_end)--];

  /* Shrink allocated memory if necessary */
  if (*free_stack_size > 2 && *free_stack_end < (*free_stack_size/4)) {
    *free_stack_size /= 2;
    *free_stack 
      = (int*) realloc(*free_stack, sizeof(int[*free_stack_size]));
    if (*free_stack == NULL)
      error("Can't reallocate memory for nltk object lookup table");
  }
  
  return retval;
}

/* Pushes the given value on the top of the free stack for the given
 * lookup table.  Resizes the memory allocated for the free stack as
 * necessary.
 */
inline void 
insert_free_stack_val (struct lookup_table* lookup, int index)
{
  int **free_stack, *free_stack_end, *free_stack_size;

  /* Some shorthand */
  free_stack = &((*lookup).free_stack);
  free_stack_end = &((*lookup).free_stack_end);
  free_stack_size = &((*lookup).free_stack_size);

  /* Grow stack and grow allocated memory if necessary */
  if (++(*free_stack_end) == *free_stack_size) {
    *free_stack_size *= *free_stack_size;
    *free_stack 
      = (int*) realloc(*free_stack, sizeof(int[*free_stack_size]));
    if (*free_stack == NULL)
      error("Can't reallocate memory for nltk object lookup table");
  }

  /* Insert the index in the stack */
  (*free_stack)[*free_stack_end] = index;
}

/* Takes a pointer to a lookup table struct and initalizes it,
 * allocating the internal hash and free value struct.
 */
inline void
init_lookup_table (struct lookup_table* lookup)
{
  (*lookup).free_stack 
    = (int*) malloc (sizeof(int[FREE_STACK_INIT_SIZE]));
  if ((*lookup).free_stack == NULL)
    error("Error allocating memory for nltk object lookup table free stack");
  (*lookup).free_stack_end = -1;
  (*lookup).free_stack_size = FREE_STACK_INIT_SIZE;
  (*lookup).current_index = INT_MIN + 1; /* INT_MIN is NA to R... */
  (*lookup).table
    = g_hash_table_new_full(g_int_hash, g_int_equal, free, NULL);
  if ((*lookup).table == NULL)
    error("Error initializing nltk object lookup table");
}

/* Inserts a new address in the lookup table, and returns the
 * associated index for later lookup. 
 */
inline int
insert_lookup_table_val (struct lookup_table* lookup, 
                         PyObject* pointer)
{
  int* index; /* GHashTable only holds pointers so we need dynam */
  index = (int*) malloc(sizeof(int)); /* -ically allocated storage */
  if (! is_free_stack_empty(lookup))
    *index = get_free_stack_val(lookup);
  else if ((*lookup).current_index == INT_MAX)
    error("The nltk object lookup table is full");
  else
    *index = (*lookup).current_index++;

  g_hash_table_insert((*lookup).table, (gpointer) index, 
                      (gpointer) pointer);

  return *index;
}

/* Grabs an address from the lookup table.  Returns NULL if the object
 * isn't found, so caller is responsible for checking.
 */
inline PyObject* 
get_lookup_table_val (struct lookup_table* lookup, int index)
{
  PyObject *result;
  result = g_hash_table_lookup((*lookup).table,
      (gconstpointer) &index);
  
  return result;
}

/* Remove an object pointer from the lookup table.  Returns a pointer
 * to the object referenced by index, or NULL if the index is not in
 * the table.
 *
 * Puts the index on the free stack if it was, indeed, in the table.
 */
inline PyObject*
remove_lookup_table_val (struct lookup_table* lookup, int index)
{
  /* Note that, while the keys are dynamically allocated, the hash
   * table takes care of freeing them when removing.
   */
  PyObject* result;
  result = get_lookup_table_val(lookup, index);
  if (result != NULL) {
    insert_free_stack_val(lookup, index);
    g_hash_table_remove((*lookup).table, (gconstpointer) &index);
  }

  return result;
}

/* Frees the lookup table memory */
inline void
finalize_lookup_table(struct lookup_table* lookup)
{
  free((*lookup).free_stack);
  g_hash_table_destroy((*lookup).table);
}

#endif /* __LOOKUP_TABLE__ */
