#ifndef __RE_H__
#define __RE_H__

#include <Python.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include "embed_nltk.h"

/* This file contains C hooks to the python re module, taking care of
 * the under-the-hood details of the RegexObject and MatchObject types
 * provided by RSNL.
 */

/**** RegexObject construction and methods ****/

/* Construct a wrapped RegexObject.
 *
 * Called directly from R using the .Call interface.
 */
inline SEXP
construct_regex_object (SEXP r_pattern, SEXP r_flagint)
{
  /* Declarations */
  PyObject *pattern, *flagint, *argtuple;
  SEXP ptr;

  /* Convert args */
  pattern = r2p_string(r_pattern);
  flagint = r2p_int(r_flagint);

  /* Stick the args into a tuple to pass. */
  argtuple = PyTuple_New(2);
  PyTuple_SetItem(argtuple, 0, pattern);
  PyTuple_SetItem(argtuple, 1, flagint);

  /* Create the RegexObject and get an R pointer to it */
  ptr = construct_wrapped_python_object(python_re_pnt(), argtuple,
      "compile", destruct_python_object);

  UNPROTECT(1); // unprotect the ptr object
  return ptr;
}

/* Various RegexObject attributes */

inline SEXP
flags_RegexObject (SEXP ptr)
{
  SEXP result;
  result = get_intattr_from_extptr(ptr, "flags");
  UNPROTECT(1);
  return result;
}
  
inline SEXP
pattern_RegexObject (SEXP ptr)
{
  SEXP result;
  result = get_strattr_from_extptr(ptr, "pattern");
  UNPROTECT(1);
  return result;
}

/* Call match on a RegexObject.  This takes an external pointer to the
 * RegexObject and a single R character string (not a vector) to match
 * against.  Also takes optional (pass NA if not using) pos and
 * endpos integer arguments.  Caller is responsible for argument
 * checking.
 *
 * Called directly from R using the .Call intervace.
 */
inline SEXP
match_RegexObject (SEXP ptr, SEXP r_string, SEXP r_pos, SEXP r_endpos)
{
  /* Declarations */
  PyObject *regex_obj, *string, *pos, *endpos, *result;
  SEXP mo_ptr;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute the match method */
  string = r2p_string(r_string);
  if (INTEGER(r_pos)[0] != NA_INTEGER) {
    pos = r2p_int(r_pos);
    if (INTEGER(r_endpos)[0] != NA_INTEGER) {
      endpos = r2p_int(r_endpos);
      result = call_python_method_va(regex_obj, "match", 
          3, string, pos, endpos);
    } else {
      result = call_python_method_va(regex_obj, "match", 2, string, pos);
    }
  } else {
    result = call_python_method_va(regex_obj, "match", 1, string);
  }

  /* Wrap and return resulting MatchObject, or return Null if no match */
  if (result == Py_None)
    return R_NilValue;

  mo_ptr = wrap_python_object(result, destruct_python_object);

  UNPROTECT(1); // unprotect mo_ptr
  return mo_ptr;
}

/* Call search on a RegexObject.  This takes an external pointer to the
 * RegexObject and a single R character string (not a vector) to match
 * against.  Also takes optional (pass NA if not using) pos and
 * endpos integer arguments.  Caller is responsible for argument
 * checking.
 *
 * Called directly from R using the .Call interface.
 * TODO: This is just like match, could reduce code.
 */
inline SEXP
search_RegexObject (SEXP ptr, SEXP r_string, SEXP r_pos, SEXP r_endpos)
{
  /* Declarations */
  PyObject *regex_obj, *string, *pos, *endpos, *result;
  SEXP mo_ptr;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute the match method */
  string = r2p_string(r_string);
  if (INTEGER(r_pos)[0] != NA_INTEGER) {
    pos = r2p_int(r_pos);
    if (INTEGER(r_endpos)[0] != NA_INTEGER) {
      endpos = r2p_int(r_endpos);
      result = call_python_method_va(regex_obj, "search", 
          3, string, pos, endpos);
    } else {
      result = call_python_method_va(regex_obj, "search", 2, string, pos);
    }
  } else {
    result = call_python_method_va(regex_obj, "search", 1, string);
  }

  /* Wrap and return resulting MatchObject, or return Null if no match */
  if (result == Py_None)
    return R_NilValue;

  mo_ptr = wrap_python_object(result, destruct_python_object);

  UNPROTECT(1); // unprotect mo_ptr
  return mo_ptr;
}

/* Call split on a RegexObject.  This takes an external pointer to the
 * RegexObject, a single R character string to split (not a vector),
 * and a maxsplit int parameter.  Returns a character vector.
 *
 * Called directly from R using the .Call interface.
 */
inline SEXP
split_RegexObject (SEXP ptr, SEXP r_string, SEXP r_maxsplit)
{
  /* Declarations */
  PyObject *regex_obj, *string, *maxsplit, *result;
  SEXP r_result;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute split method */
  string = r2p_string(r_string);
  maxsplit = r2p_int(r_maxsplit);
  result = call_python_method_va(regex_obj, "split", 2, string, maxsplit);

  /* Convert back */
  r_result = p2r_string_list(result);
  Py_DECREF(result);

  UNPROTECT(1); // unprotect r_result
  return r_result;
}

/* Call search on a RegexObject.  This takes an external pointer to the
 * RegexObject and a single R character string (not a vector) to match
 * against.  Also takes optional (pass NA if not using) pos and
 * endpos integer arguments.  Caller is responsible for argument
 * checking.
 *
 * Called directly from R using the .Call interface.
 * TODO: Again, a lot of repeated code.
 */
inline SEXP
findall_RegexObject (SEXP ptr, SEXP r_string, SEXP r_pos, SEXP r_endpos)
{
  /* Declarations */
  PyObject *regex_obj, *string, *pos, *endpos, *result;
  SEXP r_result;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute the match method */
  string = r2p_string(r_string);
  if (INTEGER(r_pos)[0] != NA_INTEGER) {
    pos = r2p_int(r_pos);
    if (INTEGER(r_endpos)[0] != NA_INTEGER) {
      endpos = r2p_int(r_endpos);
      result = call_python_method_va(regex_obj, "findall", 3, string, pos,
                                   endpos);
    } else {
      result = call_python_method_va(regex_obj, "findall", 2, string, pos);
    }
  } else {
    result = call_python_method_va(regex_obj, "findall", 1, string);
  }

  /* Convert back */
  r_result = p2r_stringorstringtuple_list(result);
  Py_DECREF(result);

  UNPROTECT(1); // unprotect r_result;
  return r_result;
}

/* Call sub on a RegexObject.  This takes an external pointer to the
 * RegexObject, a single R character replacement string (not a
 * vector), a single R character string to sub into (not a vector),
 * and a count int parameter.  Returns a character string.
 *
 * Called directly from R using the .Call interface.
 */
inline SEXP
sub_RegexObject (SEXP ptr, SEXP r_repl, SEXP r_string, SEXP r_count)
{
  /* Declarations */
  PyObject *regex_obj, *repl, *string, *count, *result;
  SEXP r_result;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute split method */
  repl = r2p_string(r_repl);
  string = r2p_string(r_string);
  count = r2p_int(r_count);
  result = call_python_method_va(regex_obj, "sub", 3, repl, string, count);

  /* Convert back */
  r_result = p2r_string(result);
  Py_DECREF(result);

  UNPROTECT(1); // unprotect r_result
  return r_result;
}


/* Call subn on a RegexObject.  This takes an external pointer to the
 * RegexObject, a single R character replacement string (not a
 * vector), a single R character string to sub into (not a vector),
 * and a count int parameter.  Returns a list containing a character string
 * and an integer.
 *
 * Called directly from R using the .Call interface.
 */
inline SEXP
subn_RegexObject (SEXP ptr, SEXP r_repl, SEXP r_string, SEXP r_count)
{
  /* Declarations */
  PyObject *regex_obj, *repl, *string, *count, *result;
  SEXP r_resstr, r_resint, r_pair;

  /* Grab the RegexObject */
  regex_obj = get_pyobj_from_extptr(ptr);

  /* Convert the arguments and execute split method */
  repl = r2p_string(r_repl);
  string = r2p_string(r_string);
  count = r2p_int(r_count);
  result = call_python_method_va(regex_obj, "subn", 3, repl, string, count);

  /* Convert back */
  r_resstr = p2r_string(PyTuple_GetItem(result, 0));
  r_resint = p2r_int(PyTuple_GetItem(result, 1));
  PROTECT(r_pair = allocVector(VECSXP, 2));
  SET_VECTOR_ELT(r_pair, 0, r_resstr);
  SET_VECTOR_ELT(r_pair, 1, r_resint);
  Py_DECREF(result);

  UNPROTECT(3); // unprotect r_resstr, r_resint, and r_pair
  return r_pair;
}

/**** MatchObject methods ****/

/* See http://docs.python.org/lib/match-objects.html
 * Caller is responsible for ensuring that template is a string.
 *
 * Called directly from R with .Call
 */
inline SEXP
expand_MatchObject (SEXP ptr, SEXP r_template)
{
  /* Declarations */
  PyObject *tplate, *result, *match_obj;
  SEXP r_result;

  /* Grab the MatchObject */
  match_obj = get_pyobj_from_extptr(ptr);

  /* Convert, call, and convert back */
  tplate = r2p_string(r_template);
  result = call_python_method_va(match_obj, "expand", 1, tplate);

  r_result = p2r_string(result);
  Py_DECREF(result);  // done with python obj

  UNPROTECT(1); // unprotect r_result
  return r_result;
}

/* See http://docs.python.org/lib/match-objects.html
 * Caller is responsible for ensuring that group is a positive (or 0)
 * int
 *
 * Called directly from R with .Call
 */
inline SEXP
start_MatchObject (SEXP ptr, SEXP r_group)
{
  /* Declarations */
  PyObject *group, *result, *match_obj;
  SEXP r_result;

  /* Grab the MatchObject */
  match_obj = get_pyobj_from_extptr(ptr);

  /* Convert, call, and convert back */
  group = r2p_int(r_group);
  result = call_python_method_va(match_obj, "start", 1, group);

  r_result = p2r_int(result);
  Py_DECREF(result);  // done with python obj

  UNPROTECT(1); // unprotect r_result
  return r_result;
}

/* See http://docs.python.org/lib/match-objects.html
 * Caller is responsible for ensuring that group is a positive (or 0)
 * int
 *
 * Called directly from R with .Call
 * TODO: code dup reduction
 */
inline SEXP
end_MatchObject (SEXP ptr, SEXP r_group)
{
  /* Declarations */
  PyObject *group, *result, *match_obj;
  SEXP r_result;

  /* Grab the MatchObject */
  match_obj = get_pyobj_from_extptr(ptr);

  /* Convert, call, and convert back */
  group = r2p_int(r_group);
  result = call_python_method_va(match_obj, "end", 1, group);

  r_result = p2r_int(result);
  Py_DECREF(result);  // done with python obj

  UNPROTECT(1); // unprotect r_result
  return r_result;
}

/* See http://docs.python.org/lib/match-objects.html
 *
 * Called directly from R with .Call
 */
inline SEXP
groups_MatchObject (SEXP ptr, SEXP r_default)
{
  /* Declarations */
  PyObject *result, *match_obj, *tmp;
  int size;
  SEXP r_result, tvec;

  /* Grab the MatchObject */
  match_obj = get_pyobj_from_extptr(ptr);

  /* Call the method, using the default default */
  result = call_python_method(match_obj, "groups", NULL, false);

  /* Process result tuple and convert with appropriate defaults */
  size = (int) PyTuple_Size(result);
  PROTECT(r_result = allocVector(VECSXP, (int) size));
  for (int i = 0; i  < size; ++i) {
    tmp = PyTuple_GetItem(result, i);
    if (tmp == Py_None)
      SET_VECTOR_ELT(r_result, i, r_default);
    else {
      PROTECT(tvec = allocVector(STRSXP, 1));
      SET_STRING_ELT(tvec, 0, mkChar(PyString_AsString(tmp)));
      SET_VECTOR_ELT(r_result, i, tvec);
      UNPROTECT(1);
    }
  }
  Py_DECREF(result);

  UNPROTECT(1); // unprotect r_result
  return r_result;
}

/* See http://docs.python.org/lib/match-objects.html
 * The groups parameter should be a list containing at least one group
 * integer (0-99) and/or group string identifiers.
 *
 * Called directly from R with .Call
 */
inline SEXP
group_MatchObject (SEXP ptr, SEXP groups)
{
  /* Declarations */
  PyObject *result, *match_obj, *list, *tuple;
  int size;
  SEXP r_result, r_item;

  /* Grab the MatchObject */
  match_obj = get_pyobj_from_extptr(ptr);

  /* Create a tuple from the args */
  size = LENGTH(groups);
  tuple = PyTuple_New(size);
  for (int i = 0; i < size; ++i) {
    r_item = VECTOR_ELT(groups, i);
    if (isString(r_item)) {
      PyTuple_SetItem(tuple, i,
          PyString_FromString(CHAR(STRING_ELT(r_item, 0))));
    } else {
      PyTuple_SetItem(tuple, i, 
          PyInt_FromLong((long) INTEGER(r_item)[0]));
    }
  }

  /* Call the group method */
  result = call_python_method(match_obj, "group", tuple, true);

  /* Convert the result tuple to a list */
  r_result = p2r_string_tuple(result);
  Py_DECREF(result);

  UNPROTECT(1); // unprotect r_result;
  return r_result;
}

/* Various attribute extraction routines.  They all follow the same
 * basic form and are designed to be called directly from R, using the
 * .Call interface.
 */

inline SEXP
pos_MatchObject (SEXP ptr)
{
  SEXP result;
  result = get_intattr_from_extptr(ptr, "pos");
  UNPROTECT(1);
  return result;
}

inline SEXP
endpos_MatchObject (SEXP ptr)
{
  SEXP result;
  result = get_intattr_from_extptr(ptr, "endpos");
  UNPROTECT(1);
  return result;
}

inline SEXP
lastindex_MatchObject (SEXP ptr)
{
  SEXP result;
  result = get_intattr_from_extptr(ptr, "lastindex");

  if (result != R_NilValue)
    UNPROTECT(1);
  return result;
}

inline SEXP
lastgroup_MatchObject (SEXP ptr)
{
  SEXP result;
  result = get_strattr_from_extptr(ptr, "lastgroup");

  if (result != R_NilValue)
    UNPROTECT(1);
  return result;
}

inline SEXP
re_MatchObject (SEXP ptr)
{
  PyObject *re_obj;
  SEXP re_ptr;

  re_obj = get_pyattr_from_extptr(ptr, "re");
  re_ptr = wrap_python_object(re_obj, destruct_python_object);

  UNPROTECT(1);
  return re_ptr;
}

inline SEXP
string_MatchObject (SEXP ptr)
{
  SEXP result;
  result = get_strattr_from_extptr(ptr, "string");
  UNPROTECT(1);
  return result;
}


#endif /* __RE_H__ */
