#ifndef __TAGGER_H__
#define __TAGGER_H__

/* Contains functions that work on all taggers.  Basically these are
 * wrappers to methods shared by all nltk taggers, most notably tag.
 *
 * Also contains some tagger-specific utility functions.
 */

#include <Python.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include "embed_nltk.h"

/* A helper function for grabbing a tagger from an external pointer.
 * TODO specific error checking to make sure tagger is a tagger type.
 */
inline PyObject*
get_tagger (SEXP ptr)
{
  PyObject* tagger = get_pyobj_from_extptr(ptr);
  return tagger;
}

/* Tags a chunk of text using any tagger.  Takes an external pointer
 * to a tagger and an R string vector containing tokens and
 * returns a 2-column R character matrix, with tokens in the first
 * column and tags in the second.
 *
 * Called directly from R with the .Call interface.
 */
inline SEXP
tag (SEXP ptr, SEXP r_tokens)
{
  PyObject *tokens, *tagger, *taggedtokens, *funcname;
  SEXP r_taggedtokens;
  
  /* Convert to a python token list */
  tokens = r2p_token_list(r_tokens);

  /* Grab tagger from external pointer */
  tagger = get_tagger(ptr);
  
  /* Call the tagger's tag method */
  taggedtokens = call_python_method_va(tagger, "tag", 1, tokens);

  /* Convert back to R */
  r_taggedtokens = p2r_tagged_token_list2matrix(taggedtokens);
  Py_DECREF(taggedtokens); // done with taggedtokens

  /* Unprotect r_taggedtokens and return */
  UNPROTECT(1);
  return r_taggedtokens;
}

/* Tags a series of sentences using any tagger.  Takes an external
 * pointer to a tagger and a list of character vectors (each vector is
 * a sentence, one token per vector item) and returns a list of
 * 2-column R character matrices (one per sentence), with tokens in
 * the first column and tags in the second.
 */

/* XXX For now just let R take care of batch.tag.  See
 * R/wrappedTagger-class.R if we decide to hook into python at the
 * batch.tag level.

inline SEXP
batch_tag (SEXP ptr, SEXP r_sentences)
{
  PyObject *sentences, *tagger, *taggedsentences, *funcname;
  SEXP r_taggedsentences;

  sentences = r2p_token_sentences(r_sentences);
  tagger = get_tagger(ptr);
  taggedsentences = 
    call_python_method_va(tagger, "batch_tag", 1, sentences);
  r_taggedsentences = p2r_tagged_sentences2matrix(taggedsentences);
  Py_DECREF(taggedsentences);
  UNPROTECT(1); // r_taggedsentences
  return r_taggedsentences;
}
*/

#endif /* __TAGGER_H__ */
